
from automat import MethodicalMachine

class Door(object):
    def unlock(self):
        print("Opening the door so you can get your food.")

    def lock(self):
        print("Locking the door so you can't steal the food.")

class Light(object):
    def on(self):
        print("Need some food over here.")

    def off(self):
        print("We're good on food for now.")

class FoodSlot(object):
    """
    Automats were a popular kind of business in the 1950s and 60s; a sort of
    restaurant-sized vending machine that served cooked food out of a
    coin-operated dispenser.

    This class represents the logic associated with a single food slot.
    """

    machine = MethodicalMachine()

    def __init__(self, door, light):
        self._door = door
        self._light = light
        self.start()

    @machine.state(initial=True)
    def initial(self):
        """
        The initial state when we are constructed.

        Note that applications never see this state, because the constructor
        provides an input to transition out of it immediately.
        """

    @machine.state()
    def empty(self):
        """
        The machine is empty (and the light asking for food is on).
        """

    @machine.input()
    def start(self):
        """
        A private input, for transitioning to the initial blank state to
        'empty', making sure the door and light are properly configured.
        """

    @machine.state()
    def ready(self):
        """
        We've got some food and we're ready to serve it.
        """

    @machine.state()
    def serving(self):
        """
        The door is open, we're serving food.
        """

    @machine.input()
    def coin(self):
        """
        A coin (of the appropriate denomination) was inserted.
        """

    @machine.input()
    def food(self):
        """
        Food was prepared and inserted into the back of the machine.
        """

    @machine.output()
    def turnOnFoodLight(self):
        """
        Turn on the 'we need food' light.
        """
        self._light.on()

    @machine.output()
    def turnOffFoodLight(self):
        """
        Turn off the 'we need food' light.
        """
        self._light.off()

    @machine.output()
    def lockDoor(self):
        """
        Lock the door, we don't need food.
        """
        self._door.lock()

    @machine.output()
    def unlockDoor(self):
        """
        Unock the door, it's chow time!.
        """
        self._door.unlock()

    @machine.input()
    def closeDoor(self):
        """
        The door was closed.
        """

    initial.upon(start, enter=empty, outputs=[lockDoor, turnOnFoodLight])
    empty.upon(food, enter=ready, outputs=[turnOffFoodLight])
    ready.upon(coin, enter=serving, outputs=[unlockDoor])
    serving.upon(closeDoor, enter=empty, outputs=[lockDoor,
                                                  turnOnFoodLight])



slot = FoodSlot(Door(), Light())

if __name__ == '__main__':
    import sys
    sys.stdout.writelines(FoodSlot.machine.asDigraph())
    # raw_input("Hit enter to make some food and put it in the slot: ")
    # slot.food()
    # raw_input("Hit enter to insert a coin: ")
    # slot.coin()
    # raw_input("Hit enter to retrieve the food and close the door: ")
    # slot.closeDoor()
    # raw_input("Hit enter to make some more food: ")
    # slot.food()




# -*- test-case-name: automat._test.test_core -*-

"""
A core state-machine abstraction.

Perhaps something that could be replaced with or integrated into machinist.
"""

from itertools import chain

_NO_STATE = "<no state>"


class NoTransition(Exception):
    """
    A finite state machine in C{state} has no transition for C{symbol}.

    @param state: the finite state machine's state at the time of the
        illegal transition.

    @param symbol: the input symbol for which no transition exists.
    """

    def __init__(self, state, symbol):
        self.state = state
        self.symbol = symbol
        super(Exception, self).__init__(
            "no transition for {} in {}".format(symbol, state)
        )


class Automaton(object):
    """
    A declaration of a finite state machine.

    Note that this is not the machine itself; it is immutable.
    """

    def __init__(self):
        """
        Initialize the set of transitions and the initial state.
        """
        self._initialState = _NO_STATE
        self._transitions = set()


    @property
    def initialState(self):
        """
        Return this automaton's initial state.
        """
        return self._initialState


    @initialState.setter
    def initialState(self, state):
        """
        Set this automaton's initial state.  Raises a ValueError if
        this automaton already has an initial state.
        """

        if self._initialState is not _NO_STATE:
            raise ValueError(
                "initial state already set to {}".format(self._initialState))

        self._initialState = state


    def addTransition(self, inState, inputSymbol, outState, outputSymbols):
        """
        Add the given transition to the outputSymbol. Raise ValueError if
        there is already a transition with the same inState and inputSymbol.
        """
        # keeping self._transitions in a flat list makes addTransition
        # O(n^2), but state machines don't tend to have hundreds of
        # transitions.
        for (anInState, anInputSymbol, anOutState, _) in self._transitions:
            if (anInState == inState and anInputSymbol == inputSymbol):
                raise ValueError(
                    "already have transition from {} via {}".format(inState, inputSymbol))
        self._transitions.add(
            (inState, inputSymbol, outState, tuple(outputSymbols))
        )


    def allTransitions(self):
        """
        All transitions.
        """
        return frozenset(self._transitions)


    def inputAlphabet(self):
        """
        The full set of symbols acceptable to this automaton.
        """
        return {inputSymbol for (inState, inputSymbol, outState,
                                 outputSymbol) in self._transitions}


    def outputAlphabet(self):
        """
        The full set of symbols which can be produced by this automaton.
        """
        return set(
            chain.from_iterable(
                outputSymbols for
                (inState, inputSymbol, outState, outputSymbols)
                in self._transitions
            )
        )


    def states(self):
        """
        All valid states; "Q" in the mathematical description of a state
        machine.
        """
        return frozenset(
            chain.from_iterable(
                (inState, outState)
                for
                (inState, inputSymbol, outState, outputSymbol)
                in self._transitions
            )
        )


    def outputForInput(self, inState, inputSymbol):
        """
        A 2-tuple of (outState, outputSymbols) for inputSymbol.
        """
        for (anInState, anInputSymbol,
             outState, outputSymbols) in self._transitions:
            if (inState, inputSymbol) == (anInState, anInputSymbol):
                return (outState, list(outputSymbols))
        raise NoTransition(state=inState, symbol=inputSymbol)


class Transitioner(object):
    """
    The combination of a current state and an L{Automaton}.
    """

    def __init__(self, automaton, initialState):
        self._automaton = automaton
        self._state = initialState
        self._tracer = None

    def setTrace(self, tracer):
        self._tracer = tracer

    def transition(self, inputSymbol):
        """
        Transition between states, returning any outputs.
        """
        outState, outputSymbols = self._automaton.outputForInput(self._state,
                                                                 inputSymbol)
        outTracer = None
        if self._tracer:
            outTracer = self._tracer(self._state._name(),
                                     inputSymbol._name(),
                                     outState._name())
        self._state = outState
        return (outputSymbols, outTracer)



from operator import itemgetter

from automat import MethodicalMachine


class LightSwitch(object):
    machine = MethodicalMachine()

    @machine.state(serialized="on")
    def on_state(self):
        "the switch is on"
    @machine.state(serialized="off", initial=True)
    def off_state(self):
        "the switch is off"
    @machine.input()
    def flip(self):
        "flip the switch"
    on_state.upon(flip, enter=off_state, outputs=[])
    off_state.upon(flip, enter=on_state, outputs=[])

    @machine.input()
    def query_power(self):
        "return True if powered, False otherwise"
    @machine.output()
    def _is_powered(self):
        return True

    @machine.output()
    def _not_powered(self):
        return False
    on_state.upon(query_power, enter=on_state, outputs=[_is_powered],
                  collector=itemgetter(0))
    off_state.upon(query_power, enter=off_state, outputs=[_not_powered],
                   collector=itemgetter(0))

    @machine.serializer()
    def save(self, state):
        return {"is-it-on": state}

    @machine.unserializer()
    def _restore(self, blob):
        return blob["is-it-on"]

    @classmethod
    def from_blob(cls, blob):
        self = cls()
        self._restore(blob)
        return self


if __name__ == "__main__":
    l = LightSwitch()
    print(l.query_power())



"""
Setup file for automat
"""

from setuptools import setup, find_packages

try:
    from m2r import parse_from_file
    long_description = parse_from_file('README.md')
except(IOError, ImportError):
    print("\n\n!!! m2r not found, long_description is bad, don't upload this to PyPI !!!\n\n")
    import io
    long_description = io.open('README.md', encoding="utf-8").read()

setup(
    name='Automat',
    use_scm_version=True,
    url='https://github.com/glyph/Automat',
    description="""
    Self-service finite-state machines for the programmer on the go.
    """.strip(),
    long_description=long_description,
    packages=find_packages(exclude=[]),
    package_dir={'automat': 'automat'},
    setup_requires=[
        'setuptools-scm',
        'm2r',
    ],
    install_requires=[
        "attrs>=19.2.0",
        "six",
    ],
    extras_require={
        "visualize": ["graphviz>0.5.1",
                      "Twisted>=16.1.1"],
    },
    entry_points={
        "console_scripts": [
            "automat-visualize = automat._visualize:tool"
        ],
    },
    author='Glyph',
    author_email='glyph@twistedmatrix.com',
    include_package_data=True,
    license="MIT",
    keywords='fsm finite state machine automata',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)



from __future__ import print_function
import argparse
import sys

import graphviz

from ._discover import findMachines


def _gvquote(s):
    return '"{}"'.format(s.replace('"', r'\"'))


def _gvhtml(s):
    return '<{}>'.format(s)


def elementMaker(name, *children, **attrs):
    """
    Construct a string from the HTML element description.
    """
    formattedAttrs = ' '.join('{}={}'.format(key, _gvquote(str(value)))
                              for key, value in sorted(attrs.items()))
    formattedChildren = ''.join(children)
    return u'<{name} {attrs}>{children}</{name}>'.format(
        name=name,
        attrs=formattedAttrs,
        children=formattedChildren)


def tableMaker(inputLabel, outputLabels, port, _E=elementMaker):
    """
    Construct an HTML table to label a state transition.
    """
    colspan = {}
    if outputLabels:
        colspan['colspan'] = str(len(outputLabels))

    inputLabelCell = _E("td",
                        _E("font",
                           inputLabel,
                           face="menlo-italic"),
                        color="purple",
                        port=port,
                        **colspan)

    pointSize = {"point-size": "9"}
    outputLabelCells = [_E("td",
                           _E("font",
                              outputLabel,
                              **pointSize),
                           color="pink")
                        for outputLabel in outputLabels]

    rows = [_E("tr", inputLabelCell)]

    if outputLabels:
        rows.append(_E("tr", *outputLabelCells))

    return _E("table", *rows)


def makeDigraph(automaton, inputAsString=repr,
                outputAsString=repr,
                stateAsString=repr):
    """
    Produce a L{graphviz.Digraph} object from an automaton.
    """
    digraph = graphviz.Digraph(graph_attr={'pack': 'true',
                                           'dpi': '100'},
                               node_attr={'fontname': 'Menlo'},
                               edge_attr={'fontname': 'Menlo'})

    for state in automaton.states():
        if state is automaton.initialState:
            stateShape = "bold"
            fontName = "Menlo-Bold"
        else:
            stateShape = ""
            fontName = "Menlo"
        digraph.node(stateAsString(state),
                     fontame=fontName,
                     shape="ellipse",
                     style=stateShape,
                     color="blue")
    for n, eachTransition in enumerate(automaton.allTransitions()):
        inState, inputSymbol, outState, outputSymbols = eachTransition
        thisTransition = "t{}".format(n)
        inputLabel = inputAsString(inputSymbol)

        port = "tableport"
        table = tableMaker(inputLabel, [outputAsString(outputSymbol)
                                        for outputSymbol in outputSymbols],
                           port=port)

        digraph.node(thisTransition,
                     label=_gvhtml(table), margin="0.2", shape="none")

        digraph.edge(stateAsString(inState),
                     '{}:{}:w'.format(thisTransition, port),
                     arrowhead="none")
        digraph.edge('{}:{}:e'.format(thisTransition, port),
                     stateAsString(outState))

    return digraph


def tool(_progname=sys.argv[0],
         _argv=sys.argv[1:],
         _syspath=sys.path,
         _findMachines=findMachines,
         _print=print):
    """
    Entry point for command line utility.
    """

    DESCRIPTION = """
    Visualize automat.MethodicalMachines as graphviz graphs.
    """
    EPILOG = """
    You must have the graphviz tool suite installed.  Please visit
    http://www.graphviz.org for more information.
    """
    if _syspath[0]:
        _syspath.insert(0, '')
    argumentParser = argparse.ArgumentParser(
        prog=_progname,
        description=DESCRIPTION,
        epilog=EPILOG)
    argumentParser.add_argument('fqpn',
                                help="A Fully Qualified Path name"
                                " representing where to find machines.")
    argumentParser.add_argument('--quiet', '-q',
                                help="suppress output",
                                default=False,
                                action="store_true")
    argumentParser.add_argument('--dot-directory', '-d',
                                help="Where to write out .dot files.",
                                default=".automat_visualize")
    argumentParser.add_argument('--image-directory', '-i',
                                help="Where to write out image files.",
                                default=".automat_visualize")
    argumentParser.add_argument('--image-type', '-t',
                                help="The image format.",
                                choices=graphviz.FORMATS,
                                default='png')
    argumentParser.add_argument('--view', '-v',
                                help="View rendered graphs with"
                                " default image viewer",
                                default=False,
                                action="store_true")
    args = argumentParser.parse_args(_argv)

    explicitlySaveDot = (args.dot_directory
                         and (not args.image_directory
                              or args.image_directory != args.dot_directory))
    if args.quiet:
        def _print(*args):
            pass

    for fqpn, machine in _findMachines(args.fqpn):
        _print(fqpn, '...discovered')

        digraph = machine.asDigraph()

        if explicitlySaveDot:
            digraph.save(filename="{}.dot".format(fqpn),
                         directory=args.dot_directory)
            _print(fqpn, "...wrote dot into", args.dot_directory)

        if args.image_directory:
            deleteDot = not args.dot_directory or explicitlySaveDot
            digraph.format = args.image_type
            digraph.render(filename="{}.dot".format(fqpn),
                           directory=args.image_directory,
                           view=args.view,
                           cleanup=deleteDot)
            if deleteDot:
                msg = "...wrote image into"
            else:
                msg = "...wrote image and dot into"
            _print(fqpn, msg, args.image_directory)
